# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# Copyright © 2016, Continuum Analytics, Inc. All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# ----------------------------------------------------------------------------
"""The ``init`` command creates a new project."""
from __future__ import absolute_import, print_function

import os

from anaconda_project import project_ops
from anaconda_project.internal.cli.console_utils import (print_project_problems, console_ask_yes_or_no)


def init_command(project_dir, assume_yes):
    """Initialize a new project.

    Returns:
        Exit code (0 on success)
    """
    # we don't want False right now because either you specify
    # --yes or we go with the default in project_ops.create
    # (depends on whether project file already exists).
    assert assume_yes is None or assume_yes is True

    if not os.path.exists(project_dir):
        if assume_yes:
            make_directory = True
        else:
            make_directory = console_ask_yes_or_no("Create directory '%s'?" % project_dir, default=False)
    else:
        make_directory = False

    project = project_ops.create(project_dir, make_directory=make_directory, fix_problems=assume_yes)
    if print_project_problems(project):
        return 1
    else:
        print("Project configuration is in %s" % project.project_file.filename)
        return 0


def main(args):
    """Start the init command and return exit status code."""
    return init_command(args.directory, args.yes)



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# Copyright © 2016, Continuum Analytics, Inc. All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# ----------------------------------------------------------------------------
from __future__ import absolute_import, print_function

import errno
import os
import uuid


def rename_over_existing(src, dest):
    try:
        # On Windows, this will throw EEXIST, on Linux it won't.
        # on Win32 / Python 2.7 it throws OSError instead of IOError
        os.rename(src, dest)
    except (OSError, IOError) as e:
        if e.errno == errno.EEXIST:
            # Clearly this song-and-dance is not in fact atomic,
            # but if something goes wrong putting the new file in
            # place at least the backup file might still be
            # around.
            backup = dest + ".bak-" + str(uuid.uuid4())
            os.rename(dest, backup)
            try:
                os.rename(src, dest)
            except Exception as e:
                os.rename(backup, dest)
                raise e
            finally:
                try:
                    os.remove(backup)
                except Exception as e:
                    pass
        else:
            raise e



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# Copyright © 2016, Continuum Analytics, Inc. All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# ----------------------------------------------------------------------------
"""Plugins base classes and functions."""
import os
from os.path import join
from anaconda_project.project_commands import (_ArgsTransformer, ProjectCommand, _http_specs)


class ArgsTrasformerTemplate(_ArgsTransformer):
    """Template class for plugins args trasformers.

    Plugins args transformers should subclass it and redefine add_class
    to implement custom arguments mapping.
    """

    def __init__(self, command):
        """Construct an ArgTransformer for the given command.

        Args:
            command (ProjectCommand): command that maps to the ArgsTransformer
        """
        super(ArgsTrasformerTemplate, self).__init__(_http_specs)
        self.command = command

    def add_args(self, results, args):
        """Overwrite this method to add custom arguments transformation.

        It should forwarding the arguments that are custom to the
        specific command served by this trasformer (
            i.e., '--anaconda-project-host' --> 'host'
        )

        Inputs:
            - results [list(tuples)]: list of 2 element tuples (option, values):
                           * option (str): name of the option
                            * values (lst(str)): list of the values passed
                                        for the option
            - args [list]: list of the args already passed in

        Returns:
            (list) list of the transformed args (that should include args)
        """
        raise RuntimeError("not implemented")  # pragma: no cover


class CommandTemplate(ProjectCommand):
    """Represents a command from the project file."""

    # ArgsTransformer to be used to parse args before forwarding them
    # to the actual command args and shell builder (choose_args_and_shell)
    args_transformer_cls = None

    # Base command to be executed. This should be overwritten by the
    # child class. I.e. command=`python` for a hypotetical python
    # command
    command = None

    def __init__(self, name, attributes):
        """Construct a command with the given attributes.

        Args:
            name (str): name of the command
            attributes (dict): named attributes of the command
        """
        super(CommandTemplate, self).__init__(name=name, attributes=attributes)
        self._args_transformer = self.args_transformer_cls(self)

    @property
    def command_with_conda_prefix(self):
        """Full command path pointing to <conda prefix>/bin."""
        return join(os.environ['CONDA_PREFIX'], 'bin', self.command)

    def _choose_args_and_shell(self, environ, extra_args=None):
        """Prepare extra args calling class _args_trasform.transform_args."""
        extra_args = self._args_transformer.transform_args(extra_args)
        return self.choose_args_and_shell(environ, extra_args=extra_args)

    def choose_args_and_shell(self, environ, extra_args=None):
        """Overwrite this method to implement custom plugin logic."""
        raise NotImplementedError()



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# Copyright © 2016, Continuum Analytics, Inc. All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# ----------------------------------------------------------------------------
"""The plugin registry (used to locate plugins)."""
from __future__ import absolute_import, print_function

from collections import namedtuple

ServiceType = namedtuple('ServiceType', ['name', 'default_variable', 'description'])


class RequirementsRegistry(object):
    """Allows creating Requirement and Provider instances."""

    def find_requirement_by_env_var(self, env_var, options):
        """Create a requirement instance given an environment variable name.

        Args:
            env_var (str): environment variable name
            options (dict): options from the project file for this requirement

        Returns:
            instance of Requirement
        """
        from .requirement import EnvVarRequirement
        return EnvVarRequirement(registry=self, env_var=env_var, options=options)

    def can_find_requirement_by_service_type(self, service_type, env_var, options):
        """See if we can create a requirement instance given a service type.

        Args:
            service_type (str): name of the service type
            env_var (str): environment variable name
            options (dict): options from the project file for this requirement

        Returns:
            boolean
        """
        # ha. obviously not the long-term implementation
        return service_type == 'redis'

    def find_requirement_by_service_type(self, service_type, env_var, options):
        """Create a requirement instance given a service type.

        Args:
            service_type (str): name of the service type
            env_var (str): environment variable name
            options (dict): options from the project file for this requirement

        Returns:
            instance of ServiceRequirement
        """
        if 'type' not in options or options['type'] != service_type:
            options = options.copy()
            options['type'] = service_type

        # future goal will be to un-hardcode this
        if service_type == 'redis':
            assert self.can_find_requirement_by_service_type(service_type, env_var, options)
            from .requirements.redis import RedisRequirement
            return RedisRequirement(registry=self, env_var=env_var, options=options)
        else:
            assert not self.can_find_requirement_by_service_type(service_type, env_var, options)
            return None

    def list_service_types(self):
        """List known service types.

        Returns:
           iterable of ``ServiceType`` named tuples with (name,default_variable,description)
        """
        return [ServiceType(name='redis', default_variable='REDIS_URL', description='A Redis server')]

    def find_provider_by_class_name(self, class_name):
        """Look up a provider by class name.

        Args:
            class_name (str): name of the provider class

        Returns:
            an instance of the passed-in class name or None if not found
        """
        # future goal will be to un-hardcode this of course
        if class_name == 'CondaEnvProvider':
            from .providers.conda_env import CondaEnvProvider
            return CondaEnvProvider()
        elif class_name == 'RedisProvider':
            from .providers.redis import RedisProvider
            return RedisProvider()
        elif class_name == 'EnvVarProvider':
            from .provider import EnvVarProvider
            return EnvVarProvider()
        elif class_name == 'DownloadProvider':
            from .providers.download import DownloadProvider
            return DownloadProvider()
        else:
            return None



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# Copyright © 2016, Continuum Analytics, Inc. All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# ----------------------------------------------------------------------------
"""Project file loading and manipulation."""
from __future__ import absolute_import

import os
from collections import OrderedDict
import json

from anaconda_project.yaml_file import YamlFile
from anaconda_project.env_spec import EnvSpec
import anaconda_project.internal.conda_api as conda_api

try:
    # this is the conda-packaged version of ruamel.yaml which has the
    # module renamed
    import ruamel_yaml as ryaml
except ImportError:  # pragma: no cover
    # this is the upstream version
    import ruamel.yaml as ryaml  # pragma: no cover

# these are in the order we'll use them if multiple are present
possible_project_file_names = ("anaconda-project.yml", "anaconda-project.yaml", "kapsel.yml", "kapsel.yaml")

DEFAULT_PROJECT_FILENAME = possible_project_file_names[0]


def _empty_default_env_spec():
    return (EnvSpec(name="default", channels=[], conda_packages=()), )


class ProjectFile(YamlFile):
    """Represents the ``anaconda-project.yml`` file which describes the project across machines/users.

    State that's specific to a machine/user/checkout/deployment
    should instead be in ``LocalStateFile``.  ``ProjectFile``
    would normally be checked in to source control or otherwise
    act as a shared resource.

    Be careful with creating your own instance of this class,
    because you have to think about when other code might load or
    save in a way that conflicts with your loads and saves.

    """

    @classmethod
    def load_for_directory(cls, directory, default_env_specs_func=_empty_default_env_spec):
        """Load the project file from the given directory, even if it doesn't exist.

        If the directory has no project file, the loaded
        ``ProjectFile`` will be empty. It won't actually be
        created on disk unless you call ``save()``.

        If the file has syntax problems, this sets the
        ``corrupted`` and ``corrupted_error_message`` properties,
        and attempts to modify the file will raise an
        exception. If the project file has semantic problems, they
        are not detected by this class but are reported by the
        ``Project`` class.

        Args:
            directory (str): path to the project directory
            default_env_specs_func (function makes list of EnvSpec): if file is created, use these

        Returns:
            a new ``ProjectFile``

        """
        for name in possible_project_file_names:
            path = os.path.join(directory, name)
            if os.path.isfile(path):
                return ProjectFile(path)
        return ProjectFile(os.path.join(directory, DEFAULT_PROJECT_FILENAME), default_env_specs_func)

    def __init__(self, filename, default_env_specs_func=_empty_default_env_spec):
        """Construct a ``ProjectFile`` with the given filename and requirement registry.

        It's easier to use ``ProjectFile.load_for_directory()`` in most cases.

        If the file has syntax problems, this sets the
        ``corrupted`` and ``corrupted_error_message`` properties,
        and attempts to modify the file will raise an
        exception. If the project file has semantic problems, they
        are not detected by this class but are reported by the
        ``Project`` class.

        Args:
            filename (str): path to the project file
            default_env_specs_func (function makes list of EnvSpec): if file is created, use these

        """
        self._default_env_specs_func = default_env_specs_func
        super(ProjectFile, self).__init__(filename)

    def _default_content(self):
        header = (
            "This is an Anaconda project file.\n" + "\n" + "Here you can describe your project and how to run it.\n" +
            "Use `anaconda-project run` to run the project.\n" +
            "The file is in YAML format, please see http://www.yaml.org/start.html for more.\n")
        sections = OrderedDict()

        sections['name'] = ("Set the 'name' key to name your project\n")

        sections['icon'] = ("Set the 'icon' key to give your project an icon\n")

        sections['description'] = ("Set a one-sentence-or-so 'description' key with project details\n")

        sections['commands'] = ("In the commands section, list your runnable scripts, notebooks, and other code.\n" +
                                "Use `anaconda-project add-command` to add commands.\n")

        sections['variables'] = ("In the variables section, list any environment variables your code depends on.\n"
                                 "Use `anaconda-project add-variable` to add variables.\n")

        sections['services'] = (
            "In the services section, list any services that should be\n" + "available before your code runs.\n" +
            "Use `anaconda-project add-service` to add services.\n")

        sections['downloads'] = ("In the downloads section, list any URLs to download to local files\n" +
                                 "before your code runs.\n" + "Use `anaconda-project add-download` to add downloads.\n")

        sections['packages'] = ("In the packages section, list any packages that must be installed\n" +
                                "before your code runs.\n" + "Use `anaconda-project add-packages` to add packages.\n")

        sections['channels'] = ("In the channels section, list any Conda channel URLs to be searched\n" +
                                "for packages.\n" + "\n" + "For example,\n" + "\n" + "channels:\n" + "   - mychannel\n")

        sections['platforms'] = ("In the platforms section, list platforms the project should work on\n" +
                                 "Examples: \"linux-64\", \"osx-64\", \"win-64\"\n" +
                                 "Use `anaconda-project add-platforms` to add platforms.\n")

        sections['env_specs'] = (
            "You can define multiple, named environment specs.\n" + "Each inherits any global packages or channels,\n" +
            "but can have its own unique ones also.\n" +
            "Use `anaconda-project add-env-spec` to add environment specs.\n")

        assert self._default_env_specs_func is not None
        default_env_specs = self._default_env_specs_func()
        assert default_env_specs is not None

        # we make a big string and then parse it because I can't figure out the
        # ruamel.yaml API to insert comments in front of map keys.
        def comment_out(comment):
            return ("# " + "\n# ".join(comment.split("\n")) + "\n").replace("# \n", "#\n")

        to_parse = comment_out(header)
        for section_name, comment in sections.items():
            # future: this is if/else is silly, we should be
            # assigning these bodies up above when we assign the
            # comments.
            if section_name == 'name':
                default_name = os.path.basename(os.path.dirname(self.filename))
                section_body = " " + json.dumps(default_name)
            elif section_name in ('icon', 'description'):
                section_body = ""  # empty body means null, not empty string
            elif section_name in ('channels', 'packages'):
                section_body = "  []"
            elif section_name == 'platforms':
                platforms = conda_api.default_platforms_with_current()
                section_body = "  [" + ", ".join(platforms) + "]"
            else:
                section_body = "  {}"
            to_parse = to_parse + "\n#\n" + comment_out(comment) + section_name + ":\n" + section_body + "\n\n\n"

        as_json = ryaml.load(to_parse, Loader=ryaml.RoundTripLoader)

        for env_spec in default_env_specs:
            as_json['env_specs'][env_spec.name] = env_spec.to_json()

        if len(default_env_specs) == 1:
            # if there's only one env spec, keep it for name/description
            # and put the packages and channels up in the global sections
            spec_name = next(iter(as_json['env_specs']))
            spec_json = as_json['env_specs'][spec_name]

            def move_list_elements(src, dest):
                # we want to preserve the dest list object with comments
                del dest[:]
                dest.extend(src)
                del src[:]

            if 'packages' in spec_json:
                move_list_elements(spec_json['packages'], as_json['packages'])
            if 'channels' in spec_json:
                move_list_elements(spec_json['channels'], as_json['channels'])
            if 'platforms' in spec_json:
                move_list_elements(spec_json['platforms'], as_json['platforms'])

        return as_json



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# Copyright © 2016, Continuum Analytics, Inc. All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# ----------------------------------------------------------------------------
"""YAML file loading and manipulation."""
from __future__ import absolute_import, print_function

# ruamel.yaml supports round-trip preserving dict ordering,
# comments, etc., which is why we use it instead of the usual yaml
# module. Remember the project file is intended to go into source
# control.
try:
    # this is the conda-packaged version of ruamel.yaml which has the
    # module renamed
    import ruamel_yaml as ryaml
    from ruamel_yaml.error import YAMLError
    from ruamel_yaml.comments import CommentedMap
    from ruamel_yaml.comments import CommentedSeq
except ImportError:  # pragma: no cover
    # this is the upstream version
    import ruamel.yaml as ryaml  # pragma: no cover
    from ruamel.yaml.error import YAMLError  # pragma: no cover
    from ruamel.yaml.comments import CommentedMap  # pragma: no cover
    from ruamel.yaml.comments import CommentedSeq  # pragma: no cover

import codecs
import errno
import os
import sys
import uuid

from anaconda_project.internal.makedirs import makedirs_ok_if_exists
from anaconda_project.internal.rename import rename_over_existing
from anaconda_project.internal.py2_compat import is_string

# We use this in other files (to abstract over the imports above)
_YAMLError = YAMLError
_CommentedMap = CommentedMap
_CommentedSeq = CommentedSeq


def _atomic_replace(path, contents, encoding='utf-8'):
    tmp = path + ".tmp-" + str(uuid.uuid4())
    try:
        with codecs.open(tmp, 'w', encoding) as file:
            file.write(contents)
            file.flush()
            file.close()
        rename_over_existing(tmp, path)
    finally:
        try:
            os.remove(tmp)
        except (IOError, OSError):
            pass


def _load_string(contents):
    if contents.strip() == '':
        # ryaml.load below returns None for an empty file, we want
        # to return an empty dict instead.
        return {}
    else:
        # using RoundTripLoader incorporates safe_load
        # (we don't load code)
        assert issubclass(ryaml.RoundTripLoader, ryaml.constructor.SafeConstructor)
        return ryaml.load(contents, Loader=ryaml.RoundTripLoader)


def _dump_string(yaml):
    return ryaml.dump(yaml, Dumper=ryaml.RoundTripDumper)


def _save_file(yaml, filename, contents=None):
    if contents is None:
        contents = _dump_string(yaml)

    try:
        # This is to ensure we don't corrupt the file, even if ruamel.yaml is broken
        ryaml.load(contents, Loader=ryaml.RoundTripLoader)
    except YAMLError as e:  # pragma: no cover (should not happen)
        print("ruamel.yaml bug; it failed to parse a file that it generated.", file=sys.stderr)
        print("  the parse error was: " + str(e), file=sys.stderr)
        print("Generated file was:", file=sys.stderr)
        print(contents, file=sys.stderr)
        raise RuntimeError("Bug in ruamel.yaml library; failed to parse a file that it generated: " + str(e))

    if not os.path.isfile(filename):
        # might have to make the directory
        dirname = os.path.dirname(filename)
        makedirs_ok_if_exists(dirname)
    _atomic_replace(filename, contents)


def _block_style_all_nodes(yaml):
    if hasattr(yaml, 'fa'):
        yaml.fa.set_block_style()

    if isinstance(yaml, list):
        for element in yaml:
            _block_style_all_nodes(element)
    elif isinstance(yaml, dict):
        for value in yaml.values():
            _block_style_all_nodes(value)


class YamlFile(object):
    """Abstract YAML file, base class for ``ProjectFile`` and ``LocalStateFile``.

    Be careful with creating your own instance of this class,
    because you have to think about when other code might load or
    save in a way that conflicts with your loads and saves.

    """

    def __init__(self, filename):
        """Load a YamlFile with the given filename.

        Raises an exception on an IOError, but if the file is
        missing this succeeds (then creates the file if and when
        you call ``save()``).

        If the file has syntax problems, this sets the
        ``corrupted`` and ``corrupted_error_message`` properties,
        and attempts to modify the file will raise an
        exception.

        """
        self.filename = filename
        self._previous_content = ""
        self._change_count = 0
        self.load()

    def load(self):
        """Reload the file from disk, discarding any unsaved changes.

        If the file has syntax problems, this sets the
        ``corrupted`` and ``corrupted_error_message`` properties,
        and attempts to modify the file will raise an
        exception.

        Returns:
            None
        """
        self._corrupted = False
        self._corrupted_error_message = None
        self._corrupted_maybe_line = None
        self._corrupted_maybe_column = None
        self._change_count = self._change_count + 1

        try:
            with codecs.open(self.filename, 'r', 'utf-8') as file:
                contents = file.read()
            self._yaml = _load_string(contents)

            # we re-dump instead of using "contents" because
            # when loading a hand-edited file, we may reformat
            # in trivial ways because our round-tripping isn't perfect,
            # and we don't want to count those trivial reformats as
            # a reason to save.
            self._previous_content = _dump_string(self._yaml)
        except IOError as e:
            if e.errno == errno.ENOENT:
                self._yaml = None
            else:
                raise e
        except YAMLError as e:
            self._corrupted = True
            self._corrupted_error_message = str(e)
            # Not sure all this paranoia is needed
            # about whether these values really exist,
            # but hard to prove it isn't.
            mark = getattr(e, 'problem_mark', None)
            if mark is not None:
                if mark.line is not None and mark.line >= 0:
                    self._corrupted_maybe_line = mark.line
                if mark.column is not None and mark.column >= 0:
                    self._corrupted_maybe_column = mark.column
            self._yaml = None

        if self._yaml is None:
            if self._corrupted:
                # don't want to throw exceptions if people get_value()
                # so stick an empty dict in here
                self._yaml = dict()
            else:
                self._yaml = self._default_content()
                # make it pretty
                _block_style_all_nodes(self._yaml)
                if not self._save_default_content():
                    # pretend we already saved
                    self._previous_content = _dump_string(self._yaml)

    def _default_comment(self):
        return "yaml file"

    def _default_content(self):
        # ruamel.yaml returns None if you load an empty file,
        # so we have to build this ourselves
        root = CommentedMap()
        root.yaml_set_start_comment(self._default_comment())
        return root

    def _save_default_content(self):
        """Override to change whether we consider a default, unmodified file dirty."""
        return True

    def _throw_if_corrupted(self):
        if self._corrupted:
            raise ValueError("Cannot modify corrupted YAML file %s\n%s" %
                             (self.filename, self._corrupted_error_message))

    @property
    def basename(self):
        """Basename of the filename."""
        return os.path.basename(self.filename)

    @property
    def corrupted(self):
        """Get whether the file is corrupted.

        A corrupted file has a syntax error so we can't modify and save it.
        See ``corrupted_error_message`` for what's wrong with it.

        Returns:
            True if file is corrupted.
        """
        return self._corrupted

    @property
    def corrupted_error_message(self):
        """Get the error message if file is corrupted, or None if it isn't.

        Use this to display the problem if the file is corrupted.

        Returns:
            Corruption message or None.
        """
        return self._corrupted_error_message

    @property
    def corrupted_maybe_line(self):
        """Get the line number for syntax error, or None if unavailable.

        Returns:
            Corruption line or None.
        """
        return self._corrupted_maybe_line

    @property
    def corrupted_maybe_column(self):
        """Get the column for syntax error, or None if unavailable.

        Returns:
            Corruption column or None.
        """
        return self._corrupted_maybe_column

    @property
    def change_count(self):
        """Get the number of times we've resynced with the file on disk (reloaded or saved changes).

        This is used for cache invalidation. If a cached value becomes invalid whenever
        ``change_count`` increments, then the cached value will be recomputed whenever
        we save new changes or reload the file.
        """
        return self._change_count

    @property
    def has_unsaved_changes(self):
        """Get whether changes are all saved."""
        # this is a fairly expensive check
        return self._previous_content != _dump_string(self._yaml)

    def use_changes_without_saving(self):
        """Apply any in-memory changes as if we'd saved, but don't actually save.

        This is used to "try out" a change before we save it. We can load()
        to undo our changes.
        """
        self._change_count = self._change_count + 1

    def save(self):
        """Write the file to disk, only if any changes have been made.

        Raises ``IOError`` if it fails for some reason.

        Returns:
            None
        """
        self._throw_if_corrupted()

        contents = _dump_string(self._yaml)
        if contents != self._previous_content:
            _save_file(self._yaml, self.filename, contents)
            self._change_count = self._change_count + 1
            self._previous_content = contents

    @classmethod
    def _path(cls, path):
        if is_string(path):
            return (path, )
        else:
            try:
                return list(element for element in path)
            except TypeError:
                raise ValueError("YAML file path must be a string or an iterable of strings")

    def _get_dict_or_none(self, pieces):
        current = self._yaml
        for p in pieces:
            if p in current and isinstance(current[p], dict):
                current = current[p]
            else:
                return None
        return current

    def _ensure_dicts_at_path(self, pieces):
        self._throw_if_corrupted()

        current = self._yaml
        for p in pieces:
            if p not in current or not isinstance(current[p], dict):
                # It's important to use CommentedMap because it preserves
                # order.
                current[p] = CommentedMap()
                _block_style_all_nodes(current[p])

            current = current[p]
        return current

    def set_value(self, path, value):
        """Set a single value at the given path.

        Overwrites any existing value at the path.

        This method does not save the file, call ``save()`` to do that.

        Args:
            path (str or list of str): single key, or list of nested keys
            value: any YAML-compatible value type
        """
        self._throw_if_corrupted()

        path = self._path(path)
        existing = self._ensure_dicts_at_path(path[:-1])
        existing[path[-1]] = value

    def unset_value(self, path):
        """Remove a single value at the given path.

        This method does not save the file, call ``save()`` to do that.

        Args:
            path (str or list of str): single key, or list of nested keys
        """
        self._throw_if_corrupted()

        path = self._path(path)

        existing = self._get_dict_or_none(path[:-1])
        key = path[-1]
        if existing is not None and key in existing:
            del existing[key]

    def get_value(self, path, default=None):
        """Get a single value from the YAML file.

        Args:
            path (str or list of str): single key, or list of nested keys
            default: any YAML-compatible value type

        Returns:
            the value from the file or the provided default
        """
        path = self._path(path)
        existing = self._get_dict_or_none(path[:-1])
        if existing is None:
            return default
        else:
            return existing.get(path[-1], default)

    @property
    def root(self):
        """Get the outermost value from the yaml file."""
        self._throw_if_corrupted()

        return self._yaml



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# Copyright © 2016, Continuum Analytics, Inc. All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# ----------------------------------------------------------------------------
"""Public API related to providing requirements."""
from __future__ import absolute_import

# "do the right thing" for prod, dev; OR just check status, don't
# do anything.
#
# In Production mode we do try to provide requirements, but not in
# ways that are likely to be nonsense in production (such as
# starting a throwaway local database).
#
# In Development mode we provide requirements that may not be
# production-suitable, such as a throwaway local database.
#
# In Check mode, we don't do anything that would start processes
# or modify the filesystem; we just see whether requirements have
# been met already.
#

PROVIDE_MODE_PRODUCTION = "production"
PROVIDE_MODE_DEVELOPMENT = "development"
PROVIDE_MODE_CHECK = "check"

_all_provide_modes = (PROVIDE_MODE_PRODUCTION, PROVIDE_MODE_DEVELOPMENT, PROVIDE_MODE_CHECK)



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# Copyright © 2016, Continuum Analytics, Inc. All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# ----------------------------------------------------------------------------
"""Prepare a project to run."""
from __future__ import absolute_import
from __future__ import print_function

from abc import ABCMeta, abstractmethod
import os
from copy import deepcopy

from anaconda_project.internal.metaclass import with_metaclass
from anaconda_project.internal.simple_status import SimpleStatus
from anaconda_project.internal.toposort import toposort_from_dependency_info
from anaconda_project.internal import conda_api
from anaconda_project.internal.py2_compat import is_string
from anaconda_project.local_state_file import LocalStateFile
from anaconda_project.provide import (_all_provide_modes, PROVIDE_MODE_DEVELOPMENT)
from anaconda_project.requirements_registry.provider import ProvideContext
from anaconda_project.requirements_registry.requirement import Requirement, EnvVarRequirement, UserConfigOverrides
from anaconda_project.requirements_registry.requirements.conda_env import CondaEnvRequirement


def _update_environ(dest, src):
    """Overwrite ``environ`` with any additions from the prepared environ.

    Does not remove any variables from ``environ``.
    """
    # updating os.environ can be a memory leak, so we only update
    # those values that actually changed.
    for key, value in src.items():
        if key not in dest or dest[key] != value:
            dest[key] = value


class PrepareResult(with_metaclass(ABCMeta)):
    """Abstract class describing the result of preparing the project to run."""

    def __init__(self, statuses, environ, overrides, env_spec_name):
        """Construct an abstract PrepareResult."""
        self._statuses = tuple(statuses)
        self._environ = environ
        self._overrides = overrides
        self._env_spec_name = env_spec_name

    def __bool__(self):
        """True if we were successful."""
        return not self.failed

    def __nonzero__(self):
        """True if we were successful."""
        return self.__bool__()  # pragma: no cover (py2 only)

    @property
    @abstractmethod
    def failed(self):
        """True if we failed to do what this stage was intended to do.

        If ``execute()`` returned non-None, the failure may not be fatal; stages
        can continue to be executed and may resolve the issue.
        """
        pass  # pragma: no cover

    @property
    def statuses(self):
        """Get latest RequirementStatus if available.

        If we failed before we even checked statuses, this will be an empty list.
        """
        return self._statuses

    def status_for(self, env_var_or_class):
        """Get status for the given env var or class, or None if unknown."""
        for status in self.statuses:
            if is_string(env_var_or_class):
                if isinstance(status.requirement, EnvVarRequirement) and \
                   status.requirement.env_var == env_var_or_class:
                    return status
            elif isinstance(status.requirement, env_var_or_class):
                return status
        return None

    @property
    def environ(self):
        """Computed environment variables for the project.

        If ``failed`` is True, this environ dict may be unmodified
        from the original provided to the prepare function.
        """
        return self._environ

    @property
    def overrides(self):
        """Override object which was passed to prepare()."""
        return self._overrides

    @property
    def errors(self):
        """Get lines of error output."""
        raise NotImplementedError()  # pragma: no cover

    @property
    def env_spec_name(self):
        """The env spec name we used for the prepare.

        If the project was broken or the user provided bad input
        before we could ask CondaEnvRequirement for the env spec
        name, at the moment we sort of take a guess at the right
        name in order to guarantee this is never None. The
        guessing is a little bit broken. But it would be a very
        obscure scenario where it matters.
        """
        return self._env_spec_name

    @property
    def env_prefix(self):
        """The prefix of the prepared env, or None if none was created."""
        status = self.status_for(CondaEnvRequirement)
        if status is None:
            return None
        varname = status.requirement.env_var
        return self._environ.get(varname, None)


class PrepareSuccess(PrepareResult):
    """Class describing the successful result of preparing the project to run."""

    def __init__(self, statuses, command_exec_info, environ, overrides, env_spec_name):
        """Construct a PrepareSuccess indicating a successful prepare stage."""
        super(PrepareSuccess, self).__init__(statuses, environ, overrides, env_spec_name)
        self._command_exec_info = command_exec_info
        assert self.env_spec_name is not None

    @property
    def failed(self):
        """Get False for PrepareSuccess."""
        return False

    @property
    def command_exec_info(self):
        """``CommandExecInfo`` instance if available, None if not."""
        return self._command_exec_info

    @property
    def errors(self):
        """Get empty list of errors."""
        return []

    def update_environ(self, environ):
        """Overwrite ``environ`` with any additions from the prepared environ.

        Does not remove any variables from ``environ``.
        """
        _update_environ(environ, self._environ)


class PrepareFailure(PrepareResult):
    """Class describing the failed result of preparing the project to run."""

    def __init__(self, statuses, errors, environ, overrides, env_spec_name=None):
        """Construct a PrepareFailure indicating a failed prepare stage."""
        super(PrepareFailure, self).__init__(statuses, environ, overrides, env_spec_name)
        self._errors = errors

    @property
    def failed(self):
        """Get True for PrepareFailure."""
        return True

    @property
    def errors(self):
        """Get non-empty list of errors."""
        return self._errors


class ConfigurePrepareContext(object):
    """Information needed to configure a stage."""

    def __init__(self, environ, local_state_file, default_env_spec_name, overrides, statuses):
        """Construct a ConfigurePrepareContext."""
        self.environ = environ
        self.local_state_file = local_state_file
        self.default_env_spec_name = default_env_spec_name
        self.overrides = overrides
        self.statuses = statuses
        if len(statuses) > 0:
            from anaconda_project.requirements_registry.requirement import RequirementStatus
            assert isinstance(statuses[0], RequirementStatus)


class PrepareStage(with_metaclass(ABCMeta)):
    """A step in the project preparation process."""

    @property
    @abstractmethod
    def description_of_action(self):
        """Get a user-visible description of what happens if this step is executed."""
        pass  # pragma: no cover

    @property
    @abstractmethod
    def failed(self):
        """Synonym for result.failed, only available after ``execute()``."""
        pass  # pragma: no cover

    @abstractmethod
    def configure(self):
        """Get a ``ConfigurePrepareContext`` or None if no configuration is needed.

        Configuration should be done before execute().

        Returns:
          a ``ConfigurePrepareContext`` or None
        """
        pass  # pragma: no cover

    @abstractmethod
    def execute(self):
        """Run this step and return a new stage, or None if we are done or failed."""
        pass  # pragma: no cover

    @property
    @abstractmethod
    def result(self):
        """The ``PrepareResult`` (only available if ``execute()`` has been called)."""
        pass  # pragma: no cover

    @property
    @abstractmethod
    def environ(self):
        """The latest environment variables (from the result if any, otherwise the pre-execute ones)."""
        pass  # pragma: no cover

    @property
    @abstractmethod
    def overrides(self):
        """User overrides."""
        pass  # pragma: no cover

    @property
    @abstractmethod
    def statuses_before_execute(self):
        """``RequirementStatus`` list before execution.

        This list includes all known requirements and their statuses, while the list
        in the ``configure()`` context only includes those that should be configured
        prior to this stage's execution.
        """
        pass  # pragma: no cover

    @property
    @abstractmethod
    def statuses_after_execute(self):
        """``RequirementStatus`` list after execution.

        This list includes all known requirements and their statuses, as changed
        by ``execute()``. This property cannot be read prior to ``execute()``.
        """
        pass  # pragma: no cover


# This is defined to keep the same requirements from old_statuses
# in the refreshed list, even if they are missing from
# rechecked_statuses, and it does not add any new requirements
# from rechecked_statuses to the refreshed list.
def _refresh_status_list(old_statuses, rechecked_statuses):
    new_by_req = dict()
    for status in rechecked_statuses:
        new_by_req[status.requirement] = status
    updated = []
    for status in old_statuses:
        updated.append(new_by_req.get(status.requirement, status))
    return updated


class _FunctionPrepareStage(PrepareStage):
    """A stage chain where the description and the execute function are passed in to the constructor."""

    def __init__(self, environ, overrides, description, statuses, execute, config_context=None):
        assert isinstance(environ, dict)
        assert config_context is None or isinstance(config_context, ConfigurePrepareContext)
        self._environ = environ
        self._overrides = overrides
        # the execute function is supposed to set these two (via accessor)
        self._result = None
        self._statuses_after_execute = None

        self._description = description
        self._statuses_before_execute = statuses
        self._execute = execute
        self._config_context = config_context

    # def __repr__(self):
    #    return "_FunctionPrepareStage(%r)" % (self._description)

    @property
    def description_of_action(self):
        return self._description

    @property
    def failed(self):
        return self.result.failed

    def configure(self):
        return self._config_context

    def execute(self):
        return self._execute(self)

    @property
    def result(self):
        if self._result is None:
            raise RuntimeError("result property isn't available until after execute()")
        return self._result

    @property
    def environ(self):
        if self._result is None:
            return self._environ
        else:
            return self.result.environ

    @property
    def overrides(self):
        return self._overrides

    @property
    def statuses_before_execute(self):
        return self._statuses_before_execute

    @property
    def statuses_after_execute(self):
        if self._statuses_after_execute is None:
            raise RuntimeError("statuses_after_execute isn't available until after execute()")
        return self._statuses_after_execute

    def set_result(self, result, rechecked_statuses):
        assert result is not None
        self._statuses_after_execute = _refresh_status_list(self._statuses_before_execute, rechecked_statuses)
        self._result = result


class _AndThenPrepareStage(PrepareStage):
    """A stage chain which runs an ``and_then`` function after it executes successfully."""

    def __init__(self, stage, and_then):
        self._stage = stage
        self._and_then = and_then

    # def __repr__(self):
    #    return "_AndThenPrepareStage(%r, %r)" % (self._stage, self._and_then)

    @property
    def description_of_action(self):
        return self._stage.description_of_action

    @property
    def failed(self):
        return self._stage.failed

    def configure(self):
        return self._stage.configure()

    def execute(self):
        next = self._stage.execute()
        if next is None:
            if self._stage.failed:
                return None
            else:
                return self._and_then(self._stage.statuses_after_execute)
        else:
            return _AndThenPrepareStage(next, self._and_then)

    @property
    def result(self):
        return self._stage.result

    @property
    def environ(self):
        return self._stage.environ

    @property
    def overrides(self):
        return self._stage.overrides

    @property
    def statuses_before_execute(self):
        return self._stage.statuses_before_execute

    @property
    def statuses_after_execute(self):
        return self._stage.statuses_after_execute


def _after_stage_success(stage, and_then):
    """Run and_then function after stage executes successfully.

    and_then may return another stage, or None. It takes
    the current list of updated statuses as a parameter.
    """
    assert stage is not None
    return _AndThenPrepareStage(stage, and_then)


def _sort_statuses(environ, local_state, statuses, missing_vars_getter):
    def get_node_key(status):
        # If we add a Requirement that isn't an EnvVarRequirement,
        # we can simply return the requirement object here as its
        # own key I believe. But for now that doesn't happen.
        assert hasattr(status.requirement, 'env_var')
        return status.requirement.env_var

    def get_dependency_keys(status):
        config_keys = set()
        for env_var in missing_vars_getter(status):
            config_keys.add(env_var)
        return config_keys

    def can_ignore_dependency_on_key(key):
        # if a key is already in the environment, we don't have to
        # worry about it existing as a node in the graph we are
        # toposorting
        return key in environ

    return toposort_from_dependency_info(statuses, get_node_key, get_dependency_keys, can_ignore_dependency_on_key)


def _in_provide_whitelist(provide_whitelist, requirement):
    if provide_whitelist is None:
        # whitelist of None means "everything"
        return True

    for env_var_or_class_or_req in provide_whitelist:
        if isinstance(env_var_or_class_or_req, Requirement):
            if requirement is env_var_or_class_or_req:
                return True
        elif is_string(env_var_or_class_or_req):
            if isinstance(requirement, EnvVarRequirement) and requirement.env_var == env_var_or_class_or_req:
                return True
        else:
            if isinstance(requirement, env_var_or_class_or_req):
                return True
    return False


def _configure_and_provide(project, environ, local_state, statuses, all_statuses, keep_going_until_success, mode,
                           provide_whitelist, overrides, command, extra_command_args):

    default_env_spec_name = project.default_env_spec_name_for_command(command)

    def provide_stage(stage):
        def get_missing_to_provide(status):
            return status.analysis.missing_env_vars_to_provide

        sorted = _sort_statuses(environ, local_state, statuses, get_missing_to_provide)

        # we have to recheck all the statuses in case configuration happened
        rechecked = []
        for status in sorted:
            rechecked.append(status.recheck(environ, local_state, default_env_spec_name, overrides))

        errors = []
        did_any_providing = False
        results_by_status = dict()

        for status in rechecked:
            if not _in_provide_whitelist(provide_whitelist, status.requirement):
                continue
            elif status.has_been_provided:
                continue
            else:
                did_any_providing = True
                context = ProvideContext(environ, local_state, default_env_spec_name, status, mode, project.frontend)
                result = status.provider.provide(status.requirement, context)
                errors.extend(result.errors)
                results_by_status[status] = result

        if did_any_providing:
            old = rechecked
            rechecked = []
            for status in old:
                rechecked.append(status.recheck(environ,
                                                local_state,
                                                default_env_spec_name,
                                                overrides,
                                                latest_provide_result=results_by_status.get(status)))

        failed = False
        for status in rechecked:
            if not status:
                errors.append("missing requirement to run this project: {requirement.description}"
                              .format(requirement=status.requirement))
                project.frontend.error(errors[-1])
                errors.append("  {why_not}".format(why_not=status.status_description))
                project.frontend.error(errors[-1])
                failed = True

        result_statuses = _refresh_status_list(all_statuses, rechecked)

        current_env_spec_name = None
        for status in result_statuses:
            if status.env_spec_name is not None:
                # we're expecting exactly one status to set this,
                # the status from the CondaEnvRequirement. Possibly
                # we sometimes do a prepare with no CondaEnvRequirement?
                # but doing one with two wouldn't make sense afaik.
                assert current_env_spec_name is None
                current_env_spec_name = status.env_spec_name

        if failed:
            stage.set_result(
                PrepareFailure(statuses=result_statuses,
                               errors=errors,
                               environ=environ,
                               overrides=overrides,
                               env_spec_name=current_env_spec_name),
                rechecked)
            if keep_going_until_success:
                return _start_over(stage.statuses_after_execute, rechecked)
            else:
                return None
        else:
            if command is None:
                exec_info = None
            else:
                exec_info = command.exec_info_for_environment(environ, extra_args=extra_command_args)
            stage.set_result(
                PrepareSuccess(statuses=result_statuses,
                               command_exec_info=exec_info,
                               environ=environ,
                               overrides=overrides,
                               env_spec_name=current_env_spec_name),
                rechecked)
            return None

    def _start_over(updated_all_statuses, updated_statuses):
        configure_context = ConfigurePrepareContext(environ=environ,
                                                    local_state_file=local_state,
                                                    default_env_spec_name=default_env_spec_name,
                                                    overrides=overrides,
                                                    statuses=updated_statuses)
        return _FunctionPrepareStage(environ, overrides, "Set up project.", updated_all_statuses, provide_stage,
                                     configure_context)

    return _start_over(all_statuses, statuses)


def _partition_first_group_to_configure(environ, local_state, statuses):
    def get_missing_to_configure(status):
        return status.analysis.missing_env_vars_to_configure

    sorted = _sort_statuses(environ, local_state, statuses, get_missing_to_configure)

    # We want "head" to be everything up to but not including the
    # first requirement that's missing needed env vars. head
    # should then include the requirement that supplies the
    # missing env var.

    head = []
    tail = []

    sorted.reverse()  # to efficiently pop
    while sorted:
        status = sorted.pop()

        missing_vars = status.provider.missing_env_vars_to_configure(status.requirement, environ, local_state)

        if len(missing_vars) > 0:
            tail.append(status)
            break
        else:
            head.append(status)

    sorted.reverse()  # put it back in order
    tail.extend(sorted)

    return (head, tail)


def _process_requirement_statuses(project, environ, local_state, current_statuses, all_statuses,
                                  keep_going_until_success, mode, provide_whitelist, overrides, command,
                                  extra_command_args):
    (initial, remaining) = _partition_first_group_to_configure(environ, local_state, current_statuses)

    # a surprising thing here is that the "stages" from
    # _configure_and_provide() can be user-visible, so we always
    # want at least one (even when all the lists are empty), but
    # we don't want to pointlessly create two by splitting off an
    # empty list when we already have a list. So we want two
    # _configure_and_provide() only if there are two non-empty lists,
    # but we always want at least one _configure_and_provide()

    def _stages_for(statuses):
        return _configure_and_provide(project, environ, local_state, statuses, all_statuses, keep_going_until_success,
                                      mode, provide_whitelist, overrides, command, extra_command_args)

    if len(initial) > 0 and len(remaining) > 0:

        def process_remaining(updated_all_statuses):
            # get the new status for each remaining requirement
            updated = _refresh_status_list(remaining, updated_all_statuses)
            return _process_requirement_statuses(project, environ, local_state, updated, updated_all_statuses,
                                                 keep_going_until_success, mode, provide_whitelist, overrides, command,
                                                 extra_command_args)

        return _after_stage_success(_stages_for(initial), process_remaining)
    elif len(initial) > 0:
        return _stages_for(initial)
    else:
        # this branch would happen if a requirement depends on an
        # already-met requirement, right now our only dependency
        # is on the env prefix which we always unset at the start
        # of prepare, so this should be unreachable. Keeping this
        # code though because in future we might want it back.
        assert False, "This code should not be reachable, bug!"  # pragma: no cover
        return _stages_for(remaining)  # pragma: no cover (not reachable)


# this used to create requirement objects for missing reqs, but
# at the moment missing reqs aren't possible, so it's changed to
# an assertion until it's possible again.
def _assert_no_missing_env_var_requirements(project, environ, local_state, overrides, command, statuses):
    by_env_var = dict()
    for status in statuses:
        # if we add requirements with no env_var, change this to
        # skip those requirements here
        assert hasattr(status.requirement, 'env_var')
        by_env_var[status.requirement.env_var] = status.requirement

    needed_env_vars = set()
    for status in statuses:
        needed_env_vars.update(status.provider.missing_env_vars_to_configure(status.requirement, environ, local_state))
        needed_env_vars.update(status.provider.missing_env_vars_to_provide(status.requirement, environ, local_state))

    # created_anything = False

    for env_var in needed_env_vars:
        assert env_var in by_env_var

    #     if env_var not in by_env_var:
    #         created_anything = True
    #         requirement = project.plugin_registry.find_requirement_by_env_var(env_var, options=dict())
    #         statuses.append(requirement.check_status(environ,
    #                                                  local_state,
    #                                                  project.default_env_spec_name_for_command(command),
    #                                                  overrides,
    #                                                  latest_provide_result=None))

    # if created_anything:
    #     # run the whole above again to find any transitive requirements of the new providers
    #     _add_missing_env_var_requirements(project, environ, local_state, overrides, command, statuses)


def _first_stage(project, environ, local_state, statuses, keep_going_until_success, mode, provide_whitelist, overrides,
                 command, extra_command_args):
    assert 'PROJECT_DIR' in environ

    _assert_no_missing_env_var_requirements(project, environ, local_state, overrides, command, statuses)

    first_stage = _process_requirement_statuses(project, environ, local_state, statuses, statuses,
                                                keep_going_until_success, mode, provide_whitelist, overrides, command,
                                                extra_command_args)

    return first_stage


def _prepare_environ_and_overrides(project, environ=None, env_spec_name=None):
    if environ is None:
        environ = os.environ

    assert 'PATH' in environ

    # we modify a copy, which 1) makes all our changes atomic and
    # 2) minimizes memory leaks on systems that use putenv().
    #
    # On Linux, it appears we must use deepcopy (vs plain copy) or
    # we still modify os.environ somehow.
    #
    # On the Windows CI server, but NOT on my local Windows 10
    # machine, deepcopy() didn't work but adding the extra .copy()
    # fixed it. The failure mode was that changes to PATH in the
    # copy were not visible via os.environ or os.getenv('PATH'),
    # but they DID affect what subprocess.Popen was able to see,
    # so that a test which modified PATH in this environ_copy
    # would break subsequent tests (such as test_conda_api which
    # tries to run conda). Anyway, presumably this is a bug in
    # something, but I'm not sure in what. If you can remove the
    # extra copy() and still pass all tests on all platforms at
    # some point in the future, feel free to clean up this
    # hackery.
    environ_copy = deepcopy(environ.copy())

    # many requirements and providers might need this, plus
    # it's useful for scripts to find their source tree.
    environ_copy['PROJECT_DIR'] = project.directory_path

    # Save and then clear out any existing environment
    existing_env_prefix = conda_api.environ_get_prefix(environ_copy)
    conda_api.environ_delete_prefix_variables(environ_copy)

    overrides = UserConfigOverrides(env_spec_name=env_spec_name, inherited_env=existing_env_prefix)

    return (environ_copy, overrides)


def _internal_prepare_in_stages(project, environ_copy, overrides, keep_going_until_success, mode, provide_whitelist,
                                command_name, command, extra_command_args):
    assert not project.problems
    if mode not in _all_provide_modes:
        raise ValueError("invalid provide mode " + mode)

    assert not (command_name is not None and command is not None)
    assert command_name is None or command_name in project.commands
    assert overrides.env_spec_name is None or overrides.env_spec_name in project.env_specs

    if command is None:
        command = project.command_for_name(command_name)
        # at this point, "command" is only None if there are no
        # commands for this project.

    local_state = LocalStateFile.load_for_directory(project.directory_path)

    statuses = []
    for requirement in project.requirements(overrides.env_spec_name):
        status = requirement.check_status(environ_copy,
                                          local_state,
                                          project.default_env_spec_name_for_command(command),
                                          overrides,
                                          latest_provide_result=None)
        statuses.append(status)

    return _first_stage(project, environ_copy, local_state, statuses, keep_going_until_success, mode, provide_whitelist,
                        overrides, command, extra_command_args)


def prepare_in_stages(project,
                      environ=None,
                      keep_going_until_success=False,
                      mode=PROVIDE_MODE_DEVELOPMENT,
                      provide_whitelist=None,
                      env_spec_name=None,
                      command_name=None,
                      command=None,
                      extra_command_args=None):
    """Get a chain of all steps needed to get a project ready to execute.

    This function does not immediately do anything; it returns a
    ``PrepareStage`` object which can be executed. Executing each
    stage may return a new stage, or may return ``None``. If a
    stage returns ``None``, preparation is done and the ``failed``
    property of the stage indicates whether it failed.

    Executing a stage may ask the user questions, may start
    services, run scripts, load configuration, install
    packages... it can do anything. Expect side effects.

    Before calling this function, command_name must be validated
    (must be in ``project.commands``, or be ``None``) and
    ``project.problems`` must be empty.

    Args:
        project (Project): the project
        environ (dict): the environment to start from (None to use os.environ)
        keep_going_until_success (bool): keep returning new stages until all requirements are met
        mode (str): One of ``PROVIDE_MODE_PRODUCTION``, ``PROVIDE_MODE_DEVELOPMENT``, ``PROVIDE_MODE_CHECK``
        provide_whitelist (iterable of str): ONLY call provide() for the listed env vars' requirements
        env_spec_name (str): the environment spec name to require, or None for default
        command_name (str): which named command to choose from the project, None for default
        command (ProjectCommand): command object, None for default
        extra_command_args (list of str): extra args for the command we prepare

    Returns:
        The first ``PrepareStage`` in the chain of steps.

    """
    (environ_copy, overrides) = _prepare_environ_and_overrides(project, environ, env_spec_name)

    return _internal_prepare_in_stages(project,
                                       environ_copy=environ_copy,
                                       overrides=overrides,
                                       keep_going_until_success=keep_going_until_success,
                                       mode=mode,
                                       provide_whitelist=provide_whitelist,
                                       command_name=command_name,
                                       command=command,
                                       extra_command_args=extra_command_args)


def _project_problems_to_prepare_failure(project, environ, overrides, would_have_used_env_spec):
    if project.problems:
        errors = []
        for problem in project.problems:
            errors.append(problem)
            project.frontend.error(problem)
        error = "Unable to load the project."
        errors.append(error)
        project.frontend.error(error)

        return PrepareFailure(statuses=(),
                              errors=errors,
                              environ=environ,
                              overrides=overrides,
                              env_spec_name=would_have_used_env_spec)
    else:
        return None


def _prepare_failure_on_bad_command_name(project, command_name, environ, overrides, would_have_used_env_spec):
    if command_name is not None and command_name not in project.commands:
        error = ("Command name '%s' is not in %s, these names were found: %s" %
                 (command_name, project.project_file.filename, ", ".join(sorted(project.commands.keys()))))
        project.frontend.error(error)
        return PrepareFailure(statuses=(),
                              errors=[error],
                              environ=environ,
                              overrides=overrides,
                              env_spec_name=would_have_used_env_spec)
    else:
        return None


def _prepare_failure_on_bad_env_spec_name(project, env_spec_name, environ, overrides, would_have_used_env_spec):
    if env_spec_name is not None and env_spec_name not in project.env_specs:
        error = ("Environment name '%s' is not in %s, these names were found: %s" %
                 (env_spec_name, project.project_file.filename, ", ".join(sorted(project.env_specs.keys()))))
        project.frontend.error(error)
        return PrepareFailure(statuses=(),
                              errors=[error],
                              environ=environ,
                              overrides=overrides,
                              env_spec_name=would_have_used_env_spec)
    else:
        return None


def _check_prepare_prerequisites(project, env_spec_name, command_name, command, environ, overrides):
    assert not (command_name is not None and command is not None)

    if command is None:
        command = project.command_for_name(command_name)
        # at this point, "command" is only None if there are no
        # commands for this project.

        # this is sort of a hack to predict which env spec the
        # CondaEnvRequirement will report in its status, since we may
        # not get to checking its status if these prereqs fail.  I'm
        # not sure yet how to refactor to avoid duplicating this logic
        # with CondaEnvRequirement; one thing we fail to handle here
        # is if CONDA_PREFIX is already set and we are in
        # inherit_environment=true mode.
    would_have_used_env_spec = overrides.env_spec_name
    if would_have_used_env_spec is None:
        if command is None:
            would_have_used_env_spec = project.default_env_spec_name
        else:
            would_have_used_env_spec = project.default_env_spec_name_for_command(command)

    failed = _project_problems_to_prepare_failure(project, environ, overrides, would_have_used_env_spec)
    if failed is None:
        failed = _prepare_failure_on_bad_env_spec_name(project, env_spec_name, environ, overrides,
                                                       would_have_used_env_spec)
    if failed is None:
        failed = _prepare_failure_on_bad_command_name(project, command_name, environ, overrides,
                                                      would_have_used_env_spec)
    return failed


def prepare_without_interaction(project,
                                environ=None,
                                mode=PROVIDE_MODE_DEVELOPMENT,
                                provide_whitelist=None,
                                env_spec_name=None,
                                command_name=None,
                                command=None,
                                extra_command_args=None):
    """Prepare a project to run one of its commands.

    This method doesn't ask the user any questions, so the
    ``provide_mode`` lets you specify defaults suitable for a
    local workstation, a production deployment, or "check only"
    defaults ("check only" means don't do anything just check
    status).

    This method returns a result object. The result object has
    a ``failed`` property.  If the result is failed, the
    ``errors`` property has the errors.  If the result is not
    failed, the ``command_exec_info`` property has the stuff
    you need to run the project's default command, and the
    ``environ`` property has the updated environment. The
    passed-in ``environ`` is not modified in-place.

    You can update your original environment with
    ``result.update_environ()`` if you like, but it's probably
    a bad idea to modify ``os.environ`` in that way because
    the calling app won't want to have the project
    environment.

    The ``environ`` should usually be kept between
    preparations, starting out as ``os.environ`` but then
    being modified by the user.

    If the project has a non-empty ``problems`` attribute,
    this function returns the project problems inside a failed
    result. So ``project.problems`` does not need to be checked in
    advance.

    Args:
        project (Project): from the ``load_project`` method
        environ (dict): os.environ or the previously-prepared environ; not modified in-place
        mode (str): mode from ``PROVIDE_MODE_PRODUCTION``, ``PROVIDE_MODE_DEVELOPMENT``, ``PROVIDE_MODE_CHECK``
        provide_whitelist (iterable of str): ONLY call provide() for the listed env vars' requirements
        env_spec_name (str): the environment spec name to require, or None for default
        command_name (str): which named command to choose from the project, None for default
        command (ProjectCommand): command object, None for default
        extra_command_args (list): extra args to include in the returned command argv

    Returns:
        a ``PrepareResult`` instance, which has a ``failed`` flag

    """
    (environ_copy, overrides) = _prepare_environ_and_overrides(project, environ, env_spec_name)

    failure = _check_prepare_prerequisites(project, env_spec_name, command_name, command, environ_copy, overrides)
    if failure is not None:
        return failure

    stage = _internal_prepare_in_stages(project,
                                        environ_copy=environ_copy,
                                        overrides=overrides,
                                        keep_going_until_success=False,
                                        mode=mode,
                                        provide_whitelist=provide_whitelist,
                                        command_name=command_name,
                                        command=command,
                                        extra_command_args=extra_command_args)

    return prepare_execute_without_interaction(stage)


def prepare_execute_without_interaction(stage):
    """Advance through the PrepareStage without any interactivity.

    Returns:
       a ``PrepareResult`` instance
    """
    result = None
    while stage is not None:
        next_stage = stage.execute()
        result = stage.result
        if result.failed:
            break
        stage = next_stage
    return result


def unprepare(project, prepare_result, whitelist=None):
    """Attempt to clean up project-scoped resources allocated by prepare().

    This will retain any user configuration choices about how to
    provide requirements, but it stops project-scoped services.
    Global system services or other services potentially shared
    among projects will not be stopped.

    To stop a single service, use ``whitelist=["SERVICE_VARIABLE"]``.

    Args:
        project (Project): the project
        prepare_result (PrepareResult): result from the previous prepare
        whitelist (iterable of str or type): ONLY call shutdown commands for the listed env vars' requirements

    Returns:
        a ``Status`` instance
    """
    if project.problems:
        errors = []
        for problem in project.problems:
            errors.append(problem)
            project.frontend.error(problem)
        return SimpleStatus(success=False, description="Unable to load the project.", errors=errors)

    local_state_file = LocalStateFile.load_for_directory(project.directory_path)

    # note: if the prepare_result was a failure before statuses
    # were even checked, then statuses could be empty
    failed_statuses = []
    failed_requirements = []
    success_statuses = []
    for status in prepare_result.statuses:
        requirement = status.requirement
        if not _in_provide_whitelist(whitelist, requirement):
            continue

        provider = status.provider
        unprovide_status = provider.unprovide(requirement, prepare_result.environ, local_state_file,
                                              prepare_result.overrides, status)
        if not unprovide_status:
            failed_requirements.append(requirement)
            failed_statuses.append(unprovide_status)
        else:
            success_statuses.append(unprovide_status)

    if not failed_statuses:
        if len(success_statuses) > 1:
            for message in [status.status_description for status in success_statuses]:
                project.frontend.info(message)
            return SimpleStatus(success=True, description="Success.")
        elif len(success_statuses) > 0:
            return success_statuses[0]
        else:
            return SimpleStatus(success=True, description="Nothing to clean up.")
    elif len(failed_statuses) == 1:
        for error in failed_statuses[0].errors:
            project.frontend.error(error)
        return failed_statuses[0]
    else:
        all_errors = [error for status in failed_statuses for error in [status.errors + [status.status_description]]]
        all_names = sorted([req.env_var for req in failed_requirements if isinstance(req, EnvVarRequirement)])
        for error in all_errors:
            project.frontend.error(error)

        return SimpleStatus(success=False,
                            description=("Failed to clean up %s." % ", ".join(all_names)),
                            errors=all_errors)



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# Copyright © 2016, Continuum Analytics, Inc. All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# ----------------------------------------------------------------------------
"""The Status type."""
from __future__ import absolute_import

import warnings

from abc import ABCMeta, abstractmethod

from anaconda_project.internal.metaclass import with_metaclass


class Status(with_metaclass(ABCMeta)):
    """Class describing a failure or success status, with logs.

    Values of this class evaluate to True in a boolean context
    if the status is successful.

    Values of this class are immutable.

    """

    def __init__(self):
        """Construct an abstract Status."""

    @property
    @abstractmethod
    def status_description(self):
        """Get a one-line-ish description of the status."""
        pass  # pragma: no cover

    @property
    def logs(self):
        """Get logs relevant to the status."""
        warnings.warn("Provide a Frontend object instead of using status.logs", DeprecationWarning)
        return ()

    @property
    @abstractmethod
    def errors(self):
        """Get error logs relevant to the status.

        A rule of thumb for this field is that anything in here should also have been
        logged to a ``Frontend`` instance, so this field is kind of just a cache
        of error messages generated by a particular operation for the convenience
        of the caller.
        """
        pass  # pragma: no cover



# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------------
# Copyright © 2016, Continuum Analytics, Inc. All rights reserved.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# ----------------------------------------------------------------------------
"""Commands related to the environments section."""
from __future__ import absolute_import, print_function

import sys

from anaconda_project.internal.cli.project_load import load_project
from anaconda_project import project_ops
from anaconda_project.internal.cli import console_utils


def _handle_status(status, success_message=None):
    if status:
        print(status.status_description)
        if success_message is not None:
            print(success_message)
        return 0
    else:
        console_utils.print_status_errors(status)
        return 1


def add_env_spec(project_dir, name, packages, channels):
    """Add an environment with packages from specified channels to the project."""
    project = load_project(project_dir)
    status = project_ops.add_env_spec(project, name=name, packages=packages, channels=channels)
    return _handle_status(status, "Added environment {} to the project file.".format(name))


def remove_env_spec(project_dir, name):
    """Remove an environment with packages from the project."""
    project = load_project(project_dir)
    status = project_ops.remove_env_spec(project, name=name)
    return _handle_status(status, "Removed environment {} from the project file.".format(name))


def export_env_spec(project_dir, name, filename):
    """Save an environment.yml file."""
    project = load_project(project_dir)
    status = project_ops.export_env_spec(project, name=name, filename=filename)
    return _handle_status(status)


def add_packages(project, environment, packages, channels):
    """Add packages to the project."""
    project = load_project(project)
    status = project_ops.add_packages(project, env_spec_name=environment, packages=packages, channels=channels)
    package_list = ", ".join(packages)
    if environment is None:
        success_message = "Added packages to project file: %s." % (package_list)
    else:
        success_message = "Added packages to environment %s in project file: %s." % (environment, package_list)
    return _handle_status(status, success_message)


def remove_packages(project, environment, packages):
    """Remove packages from the project."""
    project = load_project(project)
    status = project_ops.remove_packages(project, env_spec_name=environment, packages=packages)
    package_list = ", ".join(packages)
    if environment is None:
        success_message = "Removed packages from project file: %s." % (package_list)
    else:
        success_message = "Removed packages from environment %s in project file: %s." % (environment, package_list)
    return _handle_status(status, success_message)


def add_platforms(project, environment, platforms):
    """Add platforms to the project."""
    project = load_project(project)
    status = project_ops.add_platforms(project, env_spec_name=environment, platforms=platforms)
    package_list = ", ".join(platforms)
    if environment is None:
        success_message = "Added platforms to project file: %s." % (package_list)
    else:
        success_message = "Added platforms to environment %s in project file: %s." % (environment, package_list)
    return _handle_status(status, success_message)


def remove_platforms(project, environment, platforms):
    """Remove platforms from the project."""
    project = load_project(project)
    status = project_ops.remove_platforms(project, env_spec_name=environment, platforms=platforms)
    package_list = ", ".join(platforms)
    if environment is None:
        success_message = "Removed platforms from project file: %s." % (package_list)
    else:
        success_message = "Removed platforms from environment %s in project file: %s." % (environment, package_list)
    return _handle_status(status, success_message)


def list_env_specs(project_dir):
    """List environments in the project."""
    project = load_project(project_dir)
    if console_utils.print_project_problems(project):
        return 1
    print("Environments for project: {}\n".format(project_dir))
    console_utils.print_names_and_descriptions(project.env_specs.values())
    return 0


def list_packages(project_dir, environment):
    """List the packages for an environment in the project."""
    project = load_project(project_dir)
    if console_utils.print_project_problems(project):
        return 1
    if environment is None:
        environment = project.default_env_spec_name
    env = project.env_specs.get(environment, None)
    if env is None:
        print("Project doesn't have an environment called '{}'".format(environment), file=sys.stderr)
        return 1
    print("Packages for environment '{}':\n".format(env.name))
    print("\n".join(sorted(env.conda_packages)), end='\n\n')
    return 0


def list_platforms(project_dir, environment):
    """List the platforms for an environment in the project."""
    project = load_project(project_dir)
    if console_utils.print_project_problems(project):
        return 1
    if environment is None:
        environment = project.default_env_spec_name
    env = project.env_specs.get(environment, None)
    if env is None:
        print("Project doesn't have an environment called '{}'".format(environment), file=sys.stderr)
        return 1
    print("Platforms for environment '{}':\n".format(env.name))
    print("\n".join(sorted(env.platforms)), end='\n\n')
    return 0


def lock(project_dir, env_spec_name):
    """Lock dependency versions."""
    project = load_project(project_dir)
    if console_utils.print_project_problems(project):
        return 1
    status = project_ops.lock(project, env_spec_name=env_spec_name)
    return _handle_status(status)


def update(project_dir, env_spec_name):
    """Update dependency versions."""
    project = load_project(project_dir)
    if console_utils.print_project_problems(project):
        return 1
    status = project_ops.update(project, env_spec_name=env_spec_name)
    return _handle_status(status)


def unlock(project_dir, env_spec_name):
    """Unlock dependency versions."""
    project = load_project(project_dir)
    if console_utils.print_project_problems(project):
        return 1
    status = project_ops.unlock(project, env_spec_name=env_spec_name)
    return _handle_status(status)


def main_add(args):
    """Start the add-environment command and return exit status code."""
    return add_env_spec(args.directory, args.name, args.packages, args.channel)


def main_remove(args):
    """Start the remove-environment command and return exit status code."""
    return remove_env_spec(args.directory, args.name)


def main_export(args):
    """Start the export env spec command and return exit status code."""
    return export_env_spec(args.directory, args.name, args.filename)


def main_add_packages(args):
    """Start the add-packages command and return exit status code."""
    return add_packages(args.directory, args.env_spec, args.packages, args.channel)


def main_remove_packages(args):
    """Start the remove-packages command and return exit status code."""
    return remove_packages(args.directory, args.env_spec, args.packages)


def main_add_platforms(args):
    """Start the add-platforms command and return exit status code."""
    return add_platforms(args.directory, args.env_spec, args.platforms)


def main_remove_platforms(args):
    """Start the remove-platforms command and return exit status code."""
    return remove_platforms(args.directory, args.env_spec, args.platforms)


def main_list_env_specs(args):
    """Start the list environments command and return exit status code."""
    return list_env_specs(args.directory)


def main_list_packages(args):
    """Start the list packages command and return exit status code."""
    return list_packages(args.directory, args.env_spec)


def main_list_platforms(args):
    """Start the list platforms command and return exit status code."""
    return list_platforms(args.directory, args.env_spec)


def main_lock(args):
    """Lock dependency versions and return exit status code."""
    return lock(args.directory, args.name)


def main_update(args):
    """Update dependency versions and return exit status code."""
    return update(args.directory, args.name)


def main_unlock(args):
    """Unlock dependency versions and return exit status code."""
    return unlock(args.directory, args.name)
